//
// kaxdemux.cpp
//
// kaxdemux

#include <streams.h>     // DirectShow (includes windows.h)

#include <initguid.h>    // declares DEFINE_GUID to declare an EXTERN_C const.
                         // Use once per project.
#include <wxdebug.h>

#include "kaxoutpin.h"
#include "kaxinpin.h"
#include "kaxdemux.h"

// worker-thread control messages
#define Streaming_Start		1
#define Streaming_Pause		2
#define Streaming_Stop		3

const bool bReadDummyElements = true;

//
// Common DLL routines and tables
//

// Object table - all com objects in this DLL
CFactoryTemplate g_Templates[]=
{   { L"Matroska Demuxer Filter"
    , &CLSID_KAXDEMUX
    , CKaxDemuxFilter::CreateInstance
    , NULL
    , &CKaxDemuxFilter::sudFilter 
    }
};
int g_cTemplates = sizeof(g_Templates)/sizeof(g_Templates[0]);

// The streams.h DLL entrypoint.
extern "C" BOOL WINAPI DllEntryPoint(HINSTANCE, ULONG, LPVOID);

// The entrypoint required by the MSVC runtimes. This is used instead
// of DllEntryPoint directly to ensure global C++ classes get initialised.
BOOL WINAPI DllMain(HANDLE hDllHandle, DWORD dwReason, LPVOID lpreserved) {

    return DllEntryPoint(reinterpret_cast<HINSTANCE>(hDllHandle), dwReason, lpreserved);
}


STDAPI DllRegisterServer()
{
	HKEY key;
	DWORD disp;
   
        // TODO : use StringFromGUID2 on CLSID instead of hardcoded GUID string

	// Identification on content :
	// HKEY_CLASSES_ROOT\MediaType\{major type}\{subtype}
	// This is the type that will be available on the source filter output pin
	// Major type : {e436eb83-524f-11ce-9f53-0020af0ba770} MEDIATYPE_Stream
	// Minor type : {E73A985A-29B1-44db-BB6A-C88DA3B9E3B2} MEDIASUBTYPE_Matroska
	if (ERROR_SUCCESS == RegCreateKeyEx(HKEY_CLASSES_ROOT,
			"Media Type\\{e436eb83-524f-11ce-9f53-0020af0ba770}\\{E73A985A-29B1-44db-BB6A-C88DA3B9E3B2}",
			0, "REG_SZ", REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &key, &disp))
	{
		// We use the asynchronous file reader (CLSID_AsyncReader) as the source filter
		char* guidSource = "{E436EBB5-524F-11CE-9F53-0020AF0BA770}";
		RegSetValueEx(key, "Source Filter", 0, REG_SZ, (CONST BYTE *) guidSource, strlen(guidSource));
   
		// The pattern use the following format : offset,cb,mask,val
		char* pattern = "0,4,,1A45DFA3,18,8,,6D6174726F736B61";
		// In the pattern we check that we have
		//   - an EBML file (offset 0, 1A45DFA3)
		//   - with matroska inside (offset 18, 6D6174726F736B61 = "matroska")
		RegSetValueEx(key, "0", 0, REG_SZ, (CONST BYTE *) pattern, strlen(pattern));
		RegCloseKey(key);
	}

	// Identification by extension
	// HKEY_CLASSES_ROOT\Media Type\Extensions\.ext
	if (ERROR_SUCCESS == RegCreateKeyEx(HKEY_CLASSES_ROOT,
			"Media Type\\Extensions\\.mkv", 0, "REG_SZ",
			REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &key, &disp))
	{
		// We use the asynchronous file reader (CLSID_AsyncReader) as the source filter
		char* guidSource = "{E436EBB5-524F-11CE-9F53-0020AF0BA770}";
		RegSetValueEx(key, "Source Filter", 0, REG_SZ, (CONST BYTE *) guidSource, strlen(guidSource));
   
		// Now we define the Mediatype (MEDIATYPE_Stream) and Subtype (MEDIASUBTYPE_Matroska)
		char* mediaType = "{e436eb83-524f-11ce-9f53-0020af0ba770}";
		RegSetValueEx(key, "Media Type", 0, REG_SZ, (CONST BYTE *) mediaType, strlen(mediaType));
   
		char* subType = "{E73A985A-29B1-44db-BB6A-C88DA3B9E3B2}";
		RegSetValueEx(key, "Subtype", 0, REG_SZ, (CONST BYTE *) subType, strlen(subType));
		RegCloseKey(key);
	}

	return AMovieDllRegisterServer2(TRUE);
}

STDAPI DllUnregisterServer()
{
	RegDeleteKey(HKEY_CLASSES_ROOT, "Media Type\\{e436eb83-524f-11ce-9f53-0020af0ba770}\\{E73A985A-29B1-44db-BB6A-C88DA3B9E3B2}");
	RegDeleteKey(HKEY_CLASSES_ROOT, "Media Type\\Extensions\\.mkv");

    return AMovieDllRegisterServer2(FALSE);
}

//
// Filter kaxdemux routines
//

const AMOVIESETUP_MEDIATYPE sudInPinTypes = {
	&MEDIATYPE_Stream,
	&MEDIASUBTYPE_Matroska
};
const AMOVIESETUP_MEDIATYPE sudOutPinTypes = {
	&MEDIATYPE_NULL,
	&MEDIASUBTYPE_NULL
};

const AMOVIESETUP_PIN psudPins[] = {
	{ L"Input",
	FALSE, FALSE, FALSE, FALSE,
	&CLSID_NULL,
	L"Output",
	1,
	&sudInPinTypes },
	{ L"Output",
	FALSE, TRUE, FALSE, FALSE, //TRUE, TRUE,
	&CLSID_NULL,
	L"Input",
	1,
	&sudOutPinTypes }
};

// setup data - allows the self-registration to work.
const AMOVIESETUP_FILTER CKaxDemuxFilter::sudFilter = {
    g_Templates[0].m_ClsID		// clsID
  , g_Templates[0].m_Name		// strName
  , MERIT_NORMAL				// dwMerit
  , 2							// nPins
  , psudPins					// lpPin
};

CKaxDemuxFilter::CKaxDemuxFilter(TCHAR *tszName, LPUNKNOWN punk, HRESULT *phr)
: CBaseFilter(tszName, punk, &m_FilterLock, *sudFilter.clsID)
, m_FilterLock()
, m_pStream(NULL)
, m_pKaxIO(NULL)
, m_pInPin(NULL)
, m_OutPins(NAME("OutputPins"))
, m_nPins(1)
, m_pStreamer(NULL)
, m_TimeScale(10000.0)
, m_rtStart((long)0)
{
    NOTE("CKaxDemuxFilter::CKaxDemuxFilter");

	m_dwSeekingCaps = AM_SEEKING_CanGetDuration
		/*
		| AM_SEEKING_CanSeekForwards
        | AM_SEEKING_CanSeekBackwards
        | AM_SEEKING_CanSeekAbsolute
		*/
        | AM_SEEKING_CanGetStopPos;
    m_rtStop = _I64_MAX / 2;
    m_rtDuration = m_rtStop;
    m_dRateSeeking = 1.0;
	
	for (int i=0; i<6; i++) m_pElems[i] = NULL;
}

CKaxDemuxFilter::~CKaxDemuxFilter()
{
    NOTE("CKaxDemuxFilter::~CKaxDemuxFilter");

	// SHOULD WE FREE m_pElems ???

	if(m_pStreamer)
	{
		delete m_pStreamer;
		m_pStreamer = NULL;
	}

	if (m_pInPin != NULL) delete m_pInPin;
	if (m_pKaxIO != NULL) m_pKaxIO->close();
	if (m_OutPins.GetCount() != 0) {
		POSITION pos = m_OutPins.GetHeadPosition();
		while (pos) {
			CKaxOutPin *pPin = m_OutPins.GetNext(pos);
			delete pPin;
		}
	}
}


// Provide the way for the COM support routines in <streams.h>
// to create a CKaxDemuxFilter object
CUnknown * WINAPI CKaxDemuxFilter::CreateInstance(LPUNKNOWN punk, HRESULT *phr)
{
	DbgSetModuleLevel(LOG_ERROR,5);
    DbgSetModuleLevel(LOG_TRACE,5); // comment this to remove trace
	//DbgSetModuleLevel(LOG_MEMORY,2);
	DbgSetModuleLevel(LOG_LOCKING,5);
	
    NOTE("CKaxDemuxFilter::CreateInstance");

    CUnknown *pNewObject = new CKaxDemuxFilter(NAME("kaxdemux Object"), punk, phr );
    if (pNewObject == NULL) {
        *phr = E_OUTOFMEMORY;
    }

    return pNewObject;
}

// return pointer to a pin.
CBasePin* CKaxDemuxFilter::GetPin(int n)
{
	if (n == 0) {
		if (m_pInPin == NULL) {
			HRESULT hr = S_OK;
			m_pInPin = new CKaxInPin(this, m_pLock, &hr);
			if (FAILED(hr)) return NULL;
		}

		return m_pInPin;
	}
	else {
		n--;
		if ((n < 0) || (m_OutPins.GetCount() <= n)) return NULL;
		POSITION pos = m_OutPins.GetHeadPosition();
		while (n-- > 0) m_OutPins.GetNext(pos);
		return m_OutPins.GetNext(pos);
	}
    
    return NULL;
}

int CKaxDemuxFilter::InitKaxFile(IOCallback *pIOCb) {
	bool	bTracksLoaded = false;
	int		UpperElementLevel = 0;

    NOTE("CKaxDemuxFilter::InitKaxFile");

	// initialize Element array
	for (int i=0; i<6; i++) m_pElems[i] = NULL;

	// initialize EBML Stream
	if (pIOCb == NULL) return -1;
	if (m_pStream != NULL) return -2;	// for now we dont support this
	if ((m_pStream = new EbmlStream(*pIOCb)) == NULL) return -3;
	m_pKaxIO = pIOCb;

	// is this a valid EBML file ?
	if ((m_pElems[0] = m_pStream->FindNextID(EbmlHead::ClassInfos, 0xFFFFFFFFL)) != NULL)
	{
		// Make sure we have the expected element		
		if (!(EbmlId(*m_pElems[0]) == EbmlHead::ClassInfos.GlobalId))
		{
			return -1;
		}

		// TODO : Check the ebml DocType

		m_pElems[0]->SkipData(*m_pStream, EbmlHead_Context);
		if (m_pElems[0] != NULL)
		{
			delete m_pElems[0];
			m_pElems[0] = NULL;
		}
	}
	else
		return -4;

	// find a matroska segment
	if ((m_pElems[0] = m_pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) != NULL)
		if (!(EbmlId(*m_pElems[0]) == KaxSegment::ClassInfos.GlobalId)) return -5;

	// parse all sub-elements
	m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
	while (m_pElems[1] != NULL) {
		if (UpperElementLevel != 0) break;
		
		if (EbmlId(*m_pElems[1]) == KaxTracks::ClassInfos.GlobalId) {
			// another funky level ;-)
			m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
			while (m_pElems[2] != NULL) {
				if (UpperElementLevel != 0) break;
				if (EbmlId(*m_pElems[2]) == KaxTrackEntry::ClassInfos.GlobalId) {
					wchar_t szPinName[32];					
					HRESULT	hr = S_OK;
					
					// and fill it with those values
					KaxTrackInfoStruct TrackInfo;
					memset(&TrackInfo, 0, sizeof(KaxTrackInfoStruct));
				
					InitTrack(m_pElems, UpperElementLevel, &TrackInfo);

					// create a new outpin
					switch(TrackInfo.Type)
					{
					case track_video:
						swprintf(szPinName, L"Video %02d", this->m_nPins); break;
					case track_audio:
						swprintf(szPinName, L"Audio %02d", this->m_nPins); break;
					case track_subtitle:
						swprintf(szPinName, L"Subtitle %02d", this->m_nPins); break;
					}					
					
					CKaxOutPin *pPin = new CKaxOutPin(this, GetStreamLock(), &hr, szPinName, &TrackInfo, m_TimeScale);

					// its very likely that not every format is supported (i.e. A_VORBIS)
					if (!FAILED(hr)) {// return -99 - m_nPins;
						m_nPins++;
						m_OutPins.AddTail(pPin);
					}

					if (TrackInfo.CodecID != NULL) free(TrackInfo.CodecID);
					if (TrackInfo.CodecPrivate != NULL) free(TrackInfo.CodecPrivate);
					if (TrackInfo.Type == 0x01) delete TrackInfo.pVideoInfo;
					if (TrackInfo.Type == 0x02) delete TrackInfo.pAudioInfo;
				}

				if (UpperElementLevel > 0) {
					UpperElementLevel--;
					delete m_pElems[2];
					m_pElems[2] = m_pElems[3];
					if (UpperElementLevel > 0) break;
				}
				else {
					m_pElems[2]->SkipData(*m_pStream, m_pElems[2]->Generic().Context);
					delete m_pElems[2];

					m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
				}
			}

			bTracksLoaded = true;
		}
		else if (EbmlId(*m_pElems[1]) == KaxCluster::ClassInfos.GlobalId) {
			// store position
			m_qwKaxSegStart = m_pKaxIO->getFilePointer();

			m_pTmpElems[0] = m_pElems[0];
			m_pTmpElems[1] = m_pElems[1];

			// and exit if we already parsed the TrackEntry
			m_iLevel = 1;
			m_TmpUpElLev = m_UpElLev = UpperElementLevel;

			if (bTracksLoaded) return 1;
		}
		else if (EbmlId(*m_pElems[1]) == KaxInfo::ClassInfos.GlobalId) {
			m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
			while (m_pElems[2] != NULL) {
				if (UpperElementLevel != 0) break;
				if (EbmlId(*m_pElems[2]) == KaxTimecodeScale::ClassInfos.GlobalId) {
					KaxTimecodeScale & TimecodeScale = *static_cast<KaxTimecodeScale*>(m_pElems[2]);
					TimecodeScale.ReadData(*m_pKaxIO);
					uint32 timescale = TimecodeScale;
					m_TimeScale = (double)timescale;
					m_TimeScale /= 100.0f;	// multiplicator (convert 1ns -> 100ns)
				}
				else if (EbmlId(*m_pElems[2]) == KaxDuration::ClassInfos.GlobalId) {
					KaxDuration & Duration = *static_cast<KaxDuration*>(m_pElems[2]);
					Duration.ReadData(*m_pKaxIO);
					m_rtDuration = ((float)Duration) * m_TimeScale;
					m_rtStop = m_rtDuration;
				}

				if (UpperElementLevel > 0) {
					UpperElementLevel--;
					delete m_pElems[2];
					m_pElems[2] = m_pElems[3];
					if (UpperElementLevel > 0) break;
				}
				else {
					m_pElems[2]->SkipData(*m_pStream, m_pElems[2]->Generic().Context);
					delete m_pElems[2];

					m_pElems[2] = m_pStream->FindNextElement(m_pElems[1]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
				}
			}
		}

		if (UpperElementLevel > 0) {
			UpperElementLevel--;
			delete m_pElems[1];
			m_pElems[1] = m_pElems[2];
			if (UpperElementLevel > 0) break;
		}
		else {
			m_pElems[1]->SkipData(*m_pStream, m_pElems[1]->Generic().Context);
			delete m_pElems[1];

			m_pElems[1] = m_pStream->FindNextElement(m_pElems[0]->Generic().Context, UpperElementLevel, 0xFFFFFFFFL, bReadDummyElements);
		}
	}

	return 0;
}

// parsing KaxTrackEntry SubElements
int CKaxDemuxFilter::InitTrack(EbmlElement **elems, int &UEL, KaxTrackInfoStruct *pTrackInfos) {
	EbmlElement			*plev3, *plev4, *plev5;
	int					UpElLev = UEL;

    NOTE("CKaxDemuxFilter::InitTrack");

	plev3 = plev4 = plev5 = NULL;

	// PrE - computer video player profile
	while (1) {
		if ((plev3 = m_pStream->FindNextElement(elems[2]->Generic().Context, UpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
		if (UpElLev != 0) break;
		
		if (EbmlId(*plev3) == KaxTrackNumber::ClassInfos.GlobalId) {
			KaxTrackNumber &TrackNum = *static_cast<KaxTrackNumber*>(plev3);
			TrackNum.ReadData(*m_pKaxIO);
			pTrackInfos->Number = TrackNum;
		}
		else if (EbmlId(*plev3) == KaxTrackType::ClassInfos.GlobalId) {
			KaxTrackType &TrackType = *static_cast<KaxTrackType*>(plev3);
			TrackType.ReadData(*m_pKaxIO);
			pTrackInfos->Type = TrackType;
		}
		else if (EbmlId(*plev3) == KaxTrackFlagEnabled::ClassInfos.GlobalId) {
			KaxTrackFlagEnabled &FlagEnabled = *static_cast<KaxTrackFlagEnabled*>(plev3);
			FlagEnabled.ReadData(*m_pKaxIO);
			pTrackInfos->Enabled = FlagEnabled;
		}
		else if (EbmlId(*plev3) == KaxTrackFlagDefault::ClassInfos.GlobalId) {
			KaxTrackFlagDefault &FlagDefault = *static_cast<KaxTrackFlagDefault*>(plev3);
			FlagDefault.ReadData(*m_pKaxIO);
			pTrackInfos->Default = FlagDefault;
		}
		else if (EbmlId(*plev3) == KaxTrackFlagLacing::ClassInfos.GlobalId) {
			KaxTrackFlagLacing &FlagLacing= *static_cast<KaxTrackFlagLacing*>(plev3);
			FlagLacing.ReadData(*m_pKaxIO);
			pTrackInfos->Lacing = FlagLacing;
		}
		else if (EbmlId(*plev3) == KaxTrackMinCache::ClassInfos.GlobalId) {
			KaxTrackMinCache &MinCache = *static_cast<KaxTrackMinCache*>(plev3);
			MinCache.ReadData(*m_pKaxIO);
			pTrackInfos->MinCache = MinCache;
		}
		else if (EbmlId(*plev3) == KaxTrackMaxCache::ClassInfos.GlobalId) {
			KaxTrackMaxCache &MaxCache = *static_cast<KaxTrackMaxCache*>(plev3);
			MaxCache.ReadData(*m_pKaxIO);
			pTrackInfos->MaxCache = MaxCache;
		}
		else if (EbmlId(*plev3) == KaxTrackName::ClassInfos.GlobalId) {
			KaxTrackName &Name = *static_cast<KaxTrackName*>(plev3);
			Name.ReadData(*m_pKaxIO);
			pTrackInfos->Name = Name;
		}
		else if (EbmlId(*plev3) == KaxTrackLanguage::ClassInfos.GlobalId) {
			KaxTrackLanguage &Language = *static_cast<KaxTrackLanguage*>(plev3);
			Language.ReadData(*m_pKaxIO);
			pTrackInfos->Language = Language;
		}
		else if (EbmlId(*plev3) == KaxCodecID::ClassInfos.GlobalId) {
			KaxCodecID &CodecID = *static_cast<KaxCodecID*>(plev3);
			CodecID.ReadData(*m_pKaxIO);
			pTrackInfos->CodecID = (binary*)malloc(CodecID.GetSize() * sizeof(binary));
			memcpy(pTrackInfos->CodecID, CodecID.GetBuffer(), CodecID.GetSize());
		}
		else if (EbmlId(*plev3) == KaxCodecPrivate::ClassInfos.GlobalId) {
			KaxCodecPrivate &CodecPrivate = *static_cast<KaxCodecPrivate*>(plev3);
			CodecPrivate.ReadData(*m_pKaxIO);
			pTrackInfos->CodecPrivateLen = CodecPrivate.GetSize();
			pTrackInfos->CodecPrivate = (binary*)malloc(CodecPrivate.GetSize() * sizeof(binary));
			memcpy(pTrackInfos->CodecPrivate, CodecPrivate.GetBuffer(), CodecPrivate.GetSize());			
		}
		else if (EbmlId(*plev3) == KaxCodecName::ClassInfos.GlobalId) {
			KaxCodecName &CodecName = *static_cast<KaxCodecName*>(plev3);
			CodecName.ReadData(*m_pKaxIO);
			pTrackInfos->CodecName = CodecName;
		}
		else if (EbmlId(*plev3) == KaxCodecDecodeAll::ClassInfos.GlobalId) {
			KaxCodecDecodeAll &CodecDecodeAll = *static_cast<KaxCodecDecodeAll*>(plev3);
			CodecDecodeAll.ReadData(*m_pKaxIO);
			pTrackInfos->CodecDecodeAll = CodecDecodeAll;
		}
		else if (EbmlId(*plev3) == KaxTrackOverlay::ClassInfos.GlobalId) {
			KaxTrackOverlay &Overlay = *static_cast<KaxTrackOverlay*>(plev3);
			Overlay.ReadData(*m_pKaxIO);
			pTrackInfos->Overlay = Overlay;
		}
		else if (EbmlId(*plev3) == KaxTrackVideo::ClassInfos.GlobalId) {
			pTrackInfos->pVideoInfo = new KaxVideoTrackInfoStruct;
			memset(pTrackInfos->pVideoInfo, 0, sizeof(KaxVideoTrackInfoStruct));

			while (1) {
				if ((plev4 = m_pStream->FindNextElement(plev3->Generic().Context, UpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
				if (UpElLev != 0) break;

				if (EbmlId(*plev4) == KaxVideoFlagInterlaced::ClassInfos.GlobalId) {
					KaxVideoFlagInterlaced &Interlaced = *static_cast<KaxVideoFlagInterlaced*>(plev4);
					Interlaced.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->Interlaced = Interlaced;
				}
				else if (EbmlId(*plev4) == KaxVideoStereoMode::ClassInfos.GlobalId) {
					KaxVideoStereoMode &StereoMode = *static_cast<KaxVideoStereoMode*>(plev4);
					StereoMode.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->StereoMode = StereoMode;
				}
				else if (EbmlId(*plev4) == KaxVideoPixelWidth::ClassInfos.GlobalId) {
					KaxVideoPixelWidth &PixelWidth = *static_cast<KaxVideoPixelWidth*>(plev4);
					PixelWidth.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->PixelWidth = PixelWidth;
				}
				else if (EbmlId(*plev4) == KaxVideoPixelHeight::ClassInfos.GlobalId) {
					KaxVideoPixelHeight &PixelHeight = *static_cast<KaxVideoPixelHeight*>(plev4);
					PixelHeight.ReadData(*m_pKaxIO);
					pTrackInfos->pVideoInfo->PixelHeight = PixelHeight;
				}

				if (UpElLev > 0) {
					UpElLev--;
					delete plev4;
					plev4 = plev5;
					if (UpElLev > 0) break;
				}
				else {
					plev4->SkipData(*m_pStream, plev4->Generic().Context);
					delete plev4;
				}
			}
		}
		else if (EbmlId(*plev3) == KaxTrackAudio::ClassInfos.GlobalId) {
			pTrackInfos->pAudioInfo = new KaxAudioTrackInfoStruct;
			memset(pTrackInfos->pAudioInfo, 0, sizeof(KaxAudioTrackInfoStruct));

			while (1) {
				if ((plev4 = m_pStream->FindNextElement(plev3->Generic().Context, UpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
				if (UpElLev != 0) break;

				if (EbmlId(*plev4) == KaxAudioSamplingFreq::ClassInfos.GlobalId) {
					KaxAudioSamplingFreq &SampFreq = *static_cast<KaxAudioSamplingFreq*>(plev4);
					SampFreq.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->SamplingFreq = SampFreq;
				}
				else if (EbmlId(*plev4) == KaxAudioChannels::ClassInfos.GlobalId) {
					KaxAudioChannels &Channels = *static_cast<KaxAudioChannels*>(plev4);
					Channels.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->Channels = Channels;
				}
				else if (EbmlId(*plev4) == KaxAudioPosition::ClassInfos.GlobalId) {
					KaxAudioPosition &ChanPos = *static_cast<KaxAudioPosition*>(plev4);
					ChanPos.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->Position = ChanPos.GetBuffer();
				}
				else if (EbmlId(*plev4) == KaxAudioBitDepth::ClassInfos.GlobalId) {
					KaxAudioBitDepth &BitDepth = *static_cast<KaxAudioBitDepth*>(plev4);
					BitDepth.ReadData(*m_pKaxIO);
					pTrackInfos->pAudioInfo->BitDepth = BitDepth;
				}/*
				else if (EbmlId(*plev4) == KaxAudioSubTrack::ClassInfos.GlobalId) {
					
				}*/

				if (UpElLev > 0) {
					UpElLev--;
					delete plev4;
					plev4 = plev5;
					if (UpElLev > 0) break;
				}
				else {
					plev4->SkipData(*m_pStream, plev4->Generic().Context);
					delete plev4;
				}
			}
		}

		if (UpElLev > 0) {
			UpElLev--;
			delete plev3;
			plev3 = plev4;
			if (UpElLev > 0)
				break;
		}
		else {
			plev3->SkipData(*m_pStream, plev3->Generic().Context);
			delete plev3;
		}
	}

	elems[3] = plev3;
	elems[4] = plev4;
	elems[5] = plev5;

	UEL = UpElLev;
	return 0;
}

HRESULT CKaxDemuxFilter::ActivateOutPins(bool bActivate)
{
    NOTE1("CKaxDemuxFilter::ActivateOutPins(%s) in...",bActivate?"true":"false");

	HRESULT	 hr = S_OK;
	POSITION pos = m_OutPins.GetHeadPosition();
	while (pos) {
		CKaxOutPin *pPin = m_OutPins.GetNext(pos);
		if (pPin == NULL) continue;
		if (!pPin->IsConnected()) continue;
		pPin->SendAnyway();
		if (bActivate) hr = pPin->Active();
		else hr = pPin->Inactive();
		if (FAILED(hr)) return hr;
	}

	NOTE("CKaxDemuxFilter::ActivateOutPins out.");

	return S_OK;
}

HRESULT CKaxDemuxFilter::FlushOutPins(bool bBeginFlush)
{
    NOTE1("CKaxDemuxFilter::FlushOutPins(%s) in...",bBeginFlush?"true":"false");

	HRESULT	 hr = S_OK;
	POSITION pos = m_OutPins.GetHeadPosition();
	while (pos) {
		CKaxOutPin *pPin = m_OutPins.GetNext(pos);
		if (pPin == NULL) continue;
		if (!pPin->IsConnected()) continue;
		pPin->SendAnyway();
		if (bBeginFlush) hr = pPin->DeliverBeginFlush();
		else hr = pPin->DeliverEndFlush();
		if (FAILED(hr)) return hr;
	}

	NOTE("CKaxDemuxFilter::FlushOutPins out.");

	return S_OK;
}
/*
// nTrack must not be zero-based
HRESULT CKaxDemuxFilter::DeliverSample(int nTrack, IMediaSample *pSample, uint32 timecode, int nSample) {
	CAutoLock		StreamLock(&m_StreamLock);
	REFERENCE_TIME	tStart, tEnd;
	uint32			tTrack = nTrack;

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return E_FAIL;
	if (!pPin->IsConnected()) return E_FAIL;
/*	if (!pPin->IsRunning()) pPin->SetTrackTime(timecode);	// stream-time is zero before Run() is called

	if (timecode == 0) {
		pSample->SetSyncPoint(FALSE);
		pSample->SetDiscontinuity(TRUE);
	}
/
	// set the stream- and mediatimes
	tTrack = timecode - pPin->GetTrackTime();
	pPin->SetTrackTime(timecode);

	// 1. stream time
	pPin->GetStrTime(&tStart);
	tStart += tTrack * m_TimeScale;
	pPin->SetStrTime(&tStart);
	tStart += nSample << 1;
	tEnd = tStart + 1;
	pSample->SetTime(&tStart, &tEnd);

	// 2. media time
	pPin->GetMediaTime(&tStart);
	tStart += timecode * m_TimeScale;
	pPin->SetMediaTime(&tStart);
	tStart += nSample << 1;
	tEnd = tStart + 1;
	pSample->SetMediaTime(&tStart, &tEnd);

	return pPin->Deliver(pSample);
}
*/
HRESULT CKaxDemuxFilter::NewSegment(int nTrack, REFERENCE_TIME tStart, REFERENCE_TIME tEnd) {
	CAutoLock		StreamLock(&m_StreamLock);
	REFERENCE_TIME	t = 0;

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return E_FAIL;
	if (!pPin->IsConnected()) return E_FAIL;
	pPin->SetStrTime(&t);
	pPin->SendAnyway();
	return pPin->DeliverNewSegment(tStart, tEnd, 1.0);
}

HRESULT CKaxDemuxFilter::EndOfStream(int nTrack) {
	CAutoLock	StreamLock(&m_StreamLock);

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return E_FAIL;
	if (!pPin->IsConnected()) return E_FAIL;
	pPin->SendAnyway();
	return pPin->DeliverEndOfStream();
}
/*
HRESULT CKaxDemuxFilter::GetSampleBuff(int nTrack, IMediaSample **ppSample) {
	CAutoLock	StreamLock(&m_StreamLock);

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return E_FAIL;
	if (!pPin->IsConnected()) return E_FAIL;
	return pPin->GetDeliveryBuffer(ppSample, NULL, NULL, 0);
}

BOOL CKaxDemuxFilter::SetTrackBlockDuration(int nTrack, uint32 tSampleDuration) {
	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return FALSE;
	pPin->SetBlockDuration(tSampleDuration);
	return TRUE;
}

uint32 CKaxDemuxFilter::GetTrackBlockDuration(int nTrack) {
	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return 0;
	return pPin->GetBlockDuration();
}

void CKaxDemuxFilter::SetClusterTime(uint32 tClusterTime) {
	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (pos) {
		pPin = m_OutPins.GetNext(pos);
		pPin->SetClusterTime(tClusterTime);
	}
}
*/
BOOL CKaxDemuxFilter::DeliverBlock(int nTrack, KaxBlock *pBlock) {
	CAutoLock	StreamLock(&m_StreamLock);

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (nTrack--) pPin = m_OutPins.GetNext(pos);
	if (pPin == NULL) return FALSE;
	return pPin->DeliverBlock(pBlock);
}

HRESULT CKaxDemuxFilter::Run(REFERENCE_TIME tStart)
{
	NOTE("CKaxDemuxFilter::Run");

	CAutoLock FilterLock(&m_FilterLock);
	REFERENCE_TIME	t = 0;
	HRESULT	hr;

	hr = CBaseFilter::Run(tStart);
	if (FAILED(hr)) return hr;

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (pos) {
		pPin = m_OutPins.GetNext(pos);
		pPin->SetStrTime(&t);
		pPin->SetRunning(true);
		pPin->ResetStreamTime();
	}

	return S_OK;
}

HRESULT CKaxDemuxFilter::Pause()
{
    NOTE("CKaxDemuxFilter::Pause");
    
	CAutoLock FilterLock(&m_FilterLock);
	FILTER_STATE	State;

	ActivateOutPins(true);
	State = m_State;
	m_State = State_Paused;

	if (State == State_Paused) return S_OK;
	else if (State == State_Stopped)
	{
		for (int i=1; i<m_nPins; i++) NewSegment(i, 0, 0);

		// initialize & start worker thread
		CStrInit	strinit;

		strinit.iLevel = m_iLevel;
		strinit.iUpElLev = m_UpElLev;
		memcpy(strinit.pElems, m_pElems, sizeof(EbmlElement*)*6);
		strinit.pKaxIO = m_pKaxIO;
		strinit.pLock = GetStreamLock();
		strinit.TimeScale = m_TimeScale;
		strinit.pStream = m_pStream;
		strinit.pOutPins = &m_OutPins;
		strinit.pFilter = this;

		m_pStreamer = new CStreamer;
		if (!m_pStreamer->Create()) return S_FALSE;
		if (!m_pStreamer->CallWorker(Streaming_Start)) return S_FALSE;
		if (!m_pStreamer->CallWorker((DWORD)&strinit)) return S_FALSE;
	}
	else if (State == State_Stopped) {
		if (m_pStreamer == NULL) return S_OK;
		if (!m_pStreamer->CallWorker(Streaming_Pause)) return S_FALSE;
	}

	POSITION pos = m_OutPins.GetHeadPosition();
	CKaxOutPin *pPin;
	while (pos) {
		pPin = m_OutPins.GetNext(pos);
		pPin->SetRunning(false);
		pPin->ResetStreamTime();
	}

	return S_OK;
}

HRESULT CKaxDemuxFilter::Stop()
{
    NOTE("CKaxDemuxFilter::Stop in...");

	CAutoLock FilterLock(&m_FilterLock);

	if (m_State == State_Stopped) return S_OK;

	// shutdown worker-thread
	if (m_pStreamer == NULL) return S_OK;
	if (m_pStreamer->IsRunning())
	{
		if (!m_pStreamer->CallWorker(Streaming_Stop)) return S_FALSE;
		for (int i=1; i<m_nPins; i++) EndOfStream(i);
		ActivateOutPins(false);
	}

	m_pStreamer->Close();
	delete m_pStreamer;
	m_pStreamer = NULL;

	// reset matroska stuff (search to 00:00:00)
	/*
	m_pElems[0] = m_pTmpElems[0];
	m_pElems[1] = m_pTmpElems[1];
	m_UpElLev = m_TmpUpElLev;
	m_pKaxIO->setFilePointer(m_qwKaxSegStart);
	m_iLevel = 1;
	*/
	m_pElems[0] = 0;
	m_pElems[1] = 0;
	m_UpElLev = 0;
	m_pKaxIO->setFilePointer(0);
	m_iLevel = 0;
	
	m_State = State_Stopped;	

	NOTE("CKaxDemuxFilter::Stop out.");
	return S_OK;
}

DWORD CKaxDemuxFilter::CStreamer::ThreadProc() {
	CStrInit	*pStrInit;
	KaxBlock	*pBlock;
	DWORD		cmd;

	NOTE("CKaxDemuxFilter::CStreamer::ThreadProc in...");

	running = FALSE;
	while (1) {
		if (CheckRequest(&cmd)) {
			switch (cmd) {
				case Streaming_Start:
				{
					Reply(1);
					pStrInit = (CStrInit*)GetRequest();
					if (pStrInit == NULL) {
						Reply(0);
						return 0;
					}

					// initialize internals
					memcpy(pElems, pStrInit->pElems, sizeof(EbmlElement*)*6);
					iLevel = pStrInit->iLevel;
					iUpElLev = pStrInit->iUpElLev;
					TimeScale = pStrInit->TimeScale;
					if ((pKaxIO = pStrInit->pKaxIO) == NULL) { Reply(0); return 0; }
					if ((pLock = pStrInit->pLock) == NULL) { Reply(0); return 0; }
					if ((pStream = pStrInit->pStream) == NULL) { Reply(0); return 0; }
					if ((pOutPins = pStrInit->pOutPins) == NULL) { Reply(0); return 0; }
					if ((pFilter = pStrInit->pFilter) == NULL) { Reply(0); return 0; }
					dwClusterTime = 0;
					tStr = 0;
					Reply(1);

					running = TRUE;
					break;
				}
				case Streaming_Stop:
				{
					running = FALSE;
					Reply(1);
					NOTE("CKaxDemuxFilter::CStreamer::ThreadProc out1.");
					return 1;
				}

				case Streaming_Pause:
				{
					running = FALSE;
					Reply(1);
					break;
				}
			}
		}

		if (!running) continue;

		// deliver samples downstream (one block during each loop)

		if ((pBlock = GetNextBlock()) == NULL)
		{
			// End of file
			NOTE("CKaxDemuxFilter::CStreamer::ThreadProc before pLock->Lock();");
			pLock->Lock();
			pFilter->FlushOutPins(true);
			pFilter->FlushOutPins(false);
			for (int i=1; i<=pFilter->GetNumTracks(); i++) pFilter->EndOfStream(i);
			pFilter->ActivateOutPins(false);
			pLock->Unlock();
			NOTE("CKaxDemuxFilter::CStreamer::ThreadProc after pLock->Unlock();");			
			NOTE("CKaxDemuxFilter::CStreamer::ThreadProc break1");
			break;
		}

		if (!pFilter->DeliverBlock(pBlock->TrackNum(), pBlock))
		{
			NOTE("CKaxDemuxFilter::CStreamer::ThreadProc break2");
			break;
		}
	}

    NOTE("CKaxDemuxFilter::CStreamer::ThreadProc out2.");
	Reply(1);
	running = FALSE;
	return 1;
}

KaxBlock *CKaxDemuxFilter::CStreamer::GetNextBlock() {
	KaxCluster	*SegCluster;
	KaxBlock	*pBlock;

	switch (iLevel) {
		case -1: goto bgroup;
		case 0: break;
		case 1: goto level1;
		case 2: goto level2;
		case 3: goto level3;
		default: return NULL;
	}

	iLevel = 0;
	if ((pElems[0] = pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) == NULL) return NULL;
	while (1) {
		if (iUpElLev != 0) break;

		// segment (level 0)
		if (EbmlId(*pElems[0]) == KaxSegment::ClassInfos.GlobalId) {
			iLevel = 1;

			if ((pElems[1] = pStream->FindNextElement(pElems[0]->Generic().Context, iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
			while (1) {
				if (iUpElLev != 0) break;
level1:
				// cluster (level 1)
				if (EbmlId(*pElems[1]) == KaxCluster::ClassInfos.GlobalId) {
					SegCluster = static_cast<KaxCluster *>(pElems[1]);

					iLevel = 2;
					if ((pElems[2] = pStream->FindNextElement(pElems[1]->Generic().Context, iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
					while (1) {
						if (iUpElLev != 0) break;
level2:
						if (EbmlId(*pElems[2]) == KaxClusterTimecode::ClassInfos.GlobalId) {
							KaxClusterTimecode &ClusterTime = *static_cast<KaxClusterTimecode*>(pElems[2]);
							ClusterTime.ReadData(*pKaxIO);
							SegCluster->InitTimecode(uint32(ClusterTime));
						}
						else if (EbmlId(*pElems[2]) == KaxBlockGroup::ClassInfos.GlobalId) {

							iLevel = 3;
							if ((pElems[3] = pStream->FindNextElement(pElems[2]->Generic().Context, iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
							while (1) {
								if (iUpElLev != 0) break;
level3:
								// for now we're only interested in BLOX
								if (EbmlId(*pElems[3]) == KaxBlock::ClassInfos.GlobalId) {
									pBlock = static_cast<KaxBlock*>(pElems[3]);
									pBlock->SetParent(*((KaxCluster*)pElems[1]));
									pBlock->ReadData(*pKaxIO);
									iLevel = -1;	// means we're in blockgroup
									return pBlock;
								}
								else if (EbmlId(*pElems[3]) == KaxBlockDuration::ClassInfos.GlobalId)
								{
									// TODO : we need the BlockDuration for subtitles
									KaxBlockDuration &kaxBlockDuration = *static_cast<KaxBlockDuration*>(pElems[3]);
									kaxBlockDuration.ReadData(*pKaxIO);
									uint32 kbdur =  uint32(kaxBlockDuration) * TimeScale;
								}
bgroup:
								if (iUpElLev > 0) {
									iUpElLev--;

									if (!(EbmlId(*pElems[3]) == KaxBlock::ClassInfos.GlobalId)) delete pElems[3];		// done inside outpin's deliverblock()

									pElems[3] = pElems[4];
									if (iUpElLev > 0) break;
								}
								else {
									pElems[3]->SkipData(*pStream, pElems[3]->Generic().Context);

									if (!(EbmlId(*pElems[3]) == KaxBlock::ClassInfos.GlobalId)) delete pElems[3];		// done inside outpin's deliverblock()

									if ((pElems[3] = pStream->FindNextElement(pElems[2]->Generic().Context, iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) return NULL;		// xyx - not that good - NULL may be signal for unknown element
								}
							}
						}

						if (iUpElLev > 0) {
							iUpElLev--;
							delete pElems[2];
							pElems[2] = pElems[3];
							if (iUpElLev > 0) break;
						}
						else {
							pElems[2]->SkipData(*pStream, pElems[2]->Generic().Context);
							delete pElems[2];

							if ((pElems[2] = pStream->FindNextElement(pElems[1]->Generic().Context, iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
						}
					}
				}
/*
				else if (pElems[1]->IsDummy())//EbmlId(*pElems[1]) == DummyRawElement::ClassInfos.GlobalId)
					break;//continue;	// skip() = assert() !!!
*/		
				if (iUpElLev > 0) {
					iUpElLev--;
					delete pElems[1];
					pElems[1] = pElems[2];
					if (iUpElLev > 0) break;
				}
				else {
					pElems[1]->SkipData(*pStream, pElems[1]->Generic().Context);
					delete pElems[1];

					if ((pElems[1] = pStream->FindNextElement(pElems[0]->Generic().Context, iUpElLev, 0xFFFFFFFFL, bReadDummyElements)) == NULL) break;
				}
			}
		}
		
		if (iUpElLev > 0) {
			iUpElLev--;
			delete pElems[0];
			pElems[0] = pElems[1];
			if (iUpElLev > 0) break;
		}
		else {
			pElems[0]->SkipData(*pStream, pElems[0]->Generic().Context);
			delete pElems[0];

			if ((pElems[0] = pStream->FindNextID(KaxSegment::ClassInfos, 0xFFFFFFFFL)) == NULL) break;
		}
	}

	return NULL;
}


// ----------------------------------------------------------------------------
// IMediaSeeking implementation
// ----------------------------------------------------------------------------

// inspiration taken from CSourceSeeking (in BaseClasses\ctlutil.cpp)

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetCapabilities(DWORD *pCapabilities)
{
    CheckPointer(pCapabilities, E_POINTER);
    *pCapabilities = m_dwSeekingCaps;
    return S_OK;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::CheckCapabilities(DWORD *pCapabilities)
{
    CheckPointer(pCapabilities, E_POINTER);
	
    // make sure all requested capabilities are in our mask
    return (~m_dwSeekingCaps & *pCapabilities) ? S_FALSE : S_OK;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::IsFormatSupported(const GUID *pFormat)
{
    CheckPointer(pFormat, E_POINTER);
    // only seeking in time (REFERENCE_TIME units) is supported
    return *pFormat == TIME_FORMAT_MEDIA_TIME ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::QueryPreferredFormat(GUID *pFormat)
{
    CheckPointer(pFormat, E_POINTER);
    *pFormat = TIME_FORMAT_MEDIA_TIME;
    return S_OK;
}    
HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetTimeFormat(GUID *pFormat) 
{ 
    CheckPointer(pFormat, E_POINTER);
    *pFormat = TIME_FORMAT_MEDIA_TIME;
    return S_OK;
}
   
HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::IsUsingTimeFormat(const GUID *pFormat)
{ 
    CheckPointer(pFormat, E_POINTER);
    return *pFormat == TIME_FORMAT_MEDIA_TIME ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::SetTimeFormat(const GUID *pFormat)
{ 
    CheckPointer(pFormat, E_POINTER);
	
    // nothing to set; just check that it's TIME_FORMAT_TIME
    return *pFormat == TIME_FORMAT_MEDIA_TIME ? S_OK : E_INVALIDARG;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetDuration(LONGLONG *pDuration)
{ 
    CheckPointer(pDuration, E_POINTER);
    CAutoLock lock(m_pLock);
    *pDuration = m_rtDuration;
    return S_OK;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetStopPosition(LONGLONG *pStop)
{ 
    CheckPointer(pStop, E_POINTER);
    CAutoLock lock(m_pLock);
    *pStop = m_rtStop;
    return S_OK;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetCurrentPosition(LONGLONG *pCurrent) 
{
    // GetCurrentPosition is typically supported only in renderers and
    // not in source filters.
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::ConvertTimeFormat(LONGLONG *pTarget,
															 const GUID *pTargetFormat,
															 LONGLONG Source,
															 const GUID *pSourceFormat) 
{
    CheckPointer(pTarget, E_POINTER);
    // format guids can be null to indicate current format
	
    // since we only support TIME_FORMAT_MEDIA_TIME, we don't really
    // offer any conversions.
    if(pTargetFormat == 0 || *pTargetFormat == TIME_FORMAT_MEDIA_TIME)
    {
        if(pSourceFormat == 0 || *pSourceFormat == TIME_FORMAT_MEDIA_TIME)
        {
            *pTarget = Source;
            return S_OK;
        }
    }
	
    return E_INVALIDARG;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::SetPositions(LONGLONG *pCurrent,
														DWORD dwCurrentFlags,
														LONGLONG *pStop,
														DWORD dwStopFlags) 
{ 
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetPositions(LONGLONG *pCurrent,
														LONGLONG *pStop)
{
    if(pCurrent) {
        *pCurrent = m_rtStart;
    }
    if(pStop) {
        *pStop = m_rtStop;
    }
	
    return S_OK;;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetAvailable(LONGLONG *pEarliest,
														LONGLONG *pLatest) 
{ 
    if(pEarliest) {
        *pEarliest = 0;
    }
    if(pLatest) {
        CAutoLock lock(m_pLock);
        *pLatest = m_rtDuration;
    }
    return S_OK;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::SetRate(double dRate)
{ 
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetRate(double *pdRate) 
{ 
    CheckPointer(pdRate, E_POINTER);
    CAutoLock lock(m_pLock);
    *pdRate = m_dRateSeeking;
    return S_OK;
}

HRESULT STDMETHODCALLTYPE CKaxDemuxFilter::GetPreroll(LONGLONG *pllPreroll) 
{ 
	return E_NOTIMPL; 
}