#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

/* Stuff for File Open Dialog */
static GtkWidget *open_filesel = NULL;
GtkButton *main_button;
gchar *mpa2mka_exe, *filename, *filename_mode;

/* Start Event Handlers */
void
on_buttonStart_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window, *in_text, *out_text, *ask_dialog, *dialog_label;
  gchar *input_txt; /* for storing the input file name */
  gchar *output_txt; /* for storing the output file name */
  gchar *dialog_warning; /* Dialog text warning if the files do or don't exist */
  gboolean file_exist_result; /* for detecting if a file does/nt exist */


  /* Use the function lookup_widget() to get a pointer to the widget */
  main_window = lookup_widget (GTK_WIDGET (button), "main_window");
  /* Store the pointer in the main_button global var */
  main_button = main_window;
						
   /* Lookup the text box widgets */
  in_text = lookup_widget (main_window, "entry_inputfile");
  out_text = lookup_widget (main_window, "entry_outputfile");
  /* Retrieve the text from both text box widgets */
  input_txt = gtk_entry_get_text (GTK_ENTRY(in_text));
  output_txt = gtk_entry_get_text (GTK_ENTRY(out_text));
  
  /* Find out if the files do or don't exist */
  file_exist_result = g_file_test (input_txt, G_FILE_TEST_EXISTS);
  /* Check what the result is */
  if (!file_exist_result) {
    dialog_warning = "The input file: ";
    dialog_warning = merge_strings(dialog_warning, input_txt);
    dialog_warning = merge_strings(dialog_warning, "\nDoesn't Exist!");
    show_simple_dialog(dialog_warning,"Opps","");
    return;
  };
  file_exist_result = g_file_test (output_txt, G_FILE_TEST_EXISTS);
  /* Check what the result is */
  if (file_exist_result) {
    dialog_warning = "The output file:\n ";
    dialog_warning = merge_strings(dialog_warning, output_txt);
    dialog_warning = merge_strings(dialog_warning, " \nAlready Exists!\n\nOverwrite?");
    show_simple_dialog(dialog_warning, "Yes", "No");
    return;
  };

  /* Run mpa2mka */
  run_mpa2mka(input_txt, output_txt);
}


void
on_buttonInputFileBrowse_clicked       (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *input_file_txt;
  main_button = lookup_widget (GTK_WIDGET (button), "main_window");
  input_file_txt = lookup_widget (GTK_WIDGET (button), "entry_inputfile");
  
  filename_mode = "input";
  open_file("Locate Input File", "", gtk_entry_get_text (GTK_ENTRY(input_file_txt)));
}


void
on_buttonOutputFileBrowse_clicked      (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *file_txt;
  main_button = lookup_widget (GTK_WIDGET (button), "main_window");
  file_txt = lookup_widget (GTK_WIDGET (button), "entry_outputfile");
  if (strlen(gtk_entry_get_text (GTK_ENTRY(file_txt))) <= 2)
  {
      file_txt = lookup_widget (GTK_WIDGET (button), "entry_inputfile");
  }

  filename_mode = "output";
  open_file("Locate Output File", "", gtk_entry_get_text (GTK_ENTRY(file_txt)));
}


void
on_buttonClose_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  /* Exit the program */
  gtk_main_quit();
}

void
on_buttonShowAbout_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *windowAbout, *local_buttonAboutOk, *label_version, *logo_pixmap;
  gchar *label_version_txt;
  GdkColormap *colormap;
  GdkPixmap *gdkpixmap;
  GdkBitmap *mask;
  GtkWidget *pixmap;
  static char * mkva6_48_xpm[] = {
  "48 48 48 1",
  " 	c None",
  ".	c #FFFFFF",
  "+	c #FF2020",
  "@	c #FF0000",
  "#	c #FF3030",
  "$	c #8080FF",
  "%	c #0000FF",
  "&	c #D6D6F4",
  "*	c #FF6060",
  "=	c #FF7070",
  "-	c #2020FF",
  ";	c #5858D4",
  ">	c #FFA0A0",
  ",	c #7070FF",
  "'	c #2020C5",
  ")	c #FFE0E0",
  "!	c #FF1010",
  "~	c #FF5050",
  "{	c #FF9090",
  "]	c #FF8080",
  "^	c #A0A0FF",
  "/	c #FFD0D0",
  "(	c #FFC0C0",
  "_	c #C0C0FF",
  ":	c #D0D0FF",
  "<	c #E0E0FF",
  "[	c #FFF0F0",
  "}	c #FFB0B0",
  "|	c #9090FF",
  "1	c #FF4040",
  "2	c #FF9898",
  "3	c #4040FF",
  "4	c #F0F0FF",
  "5	c #3030FF",
  "6	c #6666D7",
  "7	c #5050FF",
  "8	c #B0B0FF",
  "9	c #6060FF",
  "0	c #1010FF",
  "a	c #FFF1F1",
  "b	c #FFEDED",
  "c	c #FFA8A8",
  "d	c #FF3D3D",
  "e	c #FFFCFC",
  "f	c #FFDDDD",
  "g	c #FF9292",
  "h	c #FF3C3C",
  "i	c #FFF8F8",
  "..............+@@@#..............$%%%$.........&",
  "..............*@@@=..............$%%-..........;",
  "..............>@@@>..............$%%,..........'",
  "..............)@@@)..............$%%$..........'",
  "...............+@!...............$%%$..........'",
  "...............~@~...............$%%$..........'",
  "...............{@]...............$%%^..........'",
  ".............../@(...............$%%_..........'",
  "................!................$%%:.........<'",
  ".........[......]................$-$..........$'",
  ".........}......)............................|%'",
  ".........][...........[..................._$-%%'",
  ".........]}.........../..................._$-%%'",
  ".........]]...........}......................|%'",
  ".........]1...........2..........|3^..........$'",
  ".........]@..........[]..........$%%_.........<'",
  ".........]@(.........}]..........$%%$..........'",
  ".........]@].........]]..........$%%$..........'",
  ".........]@1.........1]..........$%%$..........'",
  ".........]@@.........@]..........$%%$..........'",
  ".........]@@(......./@]..........$%%$.........4'",
  ".........]@@].......{@]..........$%%,..........'",
  ".........]@@1.......~@]..........$%%5..........6",
  ".........]@@@.......+@]..........$%%%^.........&",
  "............._%%%%%7.....................!@@@@@@",
  "4.............%%%%%$.....................*@@@@@@",
  "^.............3%%%%8.....................}@@@@@@",
  "9.............,%%%%<......................@@@@@@",
  "0.............8%%%0.......................~@@@@@",
  "%8............<%%%3.......................>@@@@@",
  "%9.............-%%$.............a.........[@@@@@",
  "%0.............7%%8............b(..........#@@@@",
  "%%:............$%%<............c{..........{@@@@",
  "%%,............_%0.............d*..........)@@@@",
  "%%-............4%3............e@1...........+@@@",
  "%%%:............-,............f@@...........=@@@",
  "%%%|............3|............g@@/........../@@@",
  "%%%5............$_............h@@>...........!@@",
  "%%%%<...........84...........e@@@*...........*@@",
  "%%%%|...........:............i(((/...........}@@",
  "%%%%3........................................[@@",
  "%%%%%4........................................~@",
  "%%%%%^........................................>@",
  "%%%%%7.......................[(((([...........)@",
  "%%%%%%4......................(@@@@(............#",
  "%%%%%%8......................]@@@@]............{",
  "%%%%%%9......................*@@@@*............/",
  "%%%%%%0......................#@@@@1............."};

  /* Create the About Dialog */
  windowAbout = create_windowAbout ();
  /* Connect the Ok button in the About dialog to the gtk_widget_destroy */
  local_buttonAboutOk = lookup_widget (windowAbout, "buttonAboutOk");
  gtk_signal_connect_object(GTK_OBJECT (local_buttonAboutOk), "clicked", gtk_widget_destroy, GTK_OBJECT(windowAbout));
  /* Lookup and display GTK+ & GLib versions */
  label_version = lookup_widget (windowAbout, "label_version_info");
  label_version_txt = (gchar *)malloc(strlen("Running with GTK+ %i.%i.%i GLib %i.%i.%i") + 1);
  g_sprintf(label_version_txt, "Running with GTK+ %i.%i.%i GLib %i.%i.%i", gtk_major_version, gtk_minor_version, gtk_micro_version, GLIB_MAJOR_VERSION, GLIB_MINOR_VERSION, GLIB_MICRO_VERSION);
  gtk_label_set_text(GTK_LABEL(label_version), label_version_txt);    

  /* Load logo */
  logo_pixmap = lookup_widget (windowAbout, "pixmap1");  
  colormap = gtk_widget_get_colormap (windowAbout);
  gdkpixmap = gdk_pixmap_colormap_create_from_xpm_d (NULL, colormap, &mask, NULL, mkva6_48_xpm);                                                 
  gtk_pixmap_set (logo_pixmap, gdkpixmap, mask);
  gdk_pixmap_unref (gdkpixmap);
  gdk_bitmap_unref (mask);
  
  /* Show the About Dialog */
  gtk_widget_show (windowAbout);
}

void
on_buttonAboutOk_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  /* This is already connected to the gtk_widget_destroy */
}

/* This shows the file selection dialog to open a file. */
void
open_file (gchar *file_dialog_text, gchar *pattern, gchar *target_filename)
{
  GtkWidget *ok_buttonFileDialog;
  GtkWidget *cancel_buttonFileDialog;

  /* Create file selection widget so we can use it */
  open_filesel = gtk_file_selection_new (file_dialog_text);
  /* Set the data for the file selection widget so we can look it up easily */
  gtk_object_set_data (GTK_OBJECT (open_filesel), "open_filesel", open_filesel);
  gtk_container_set_border_width (GTK_CONTAINER (open_filesel), 10);
  GTK_WINDOW (open_filesel)->type = GTK_WINDOW_TOPLEVEL;

  /* Set the file pattern */
  if (pattern != "")
  {
    gtk_file_selection_complete (open_filesel, pattern);
  }
  gtk_file_selection_set_filename (open_filesel, target_filename);
  /* Create the ok button */
  ok_buttonFileDialog = GTK_FILE_SELECTION (open_filesel)->ok_button;
  /* Set the data for the ok button widget so we can look it up easily */
  gtk_object_set_data (GTK_OBJECT (open_filesel), "ok_buttonFileDialog", ok_buttonFileDialog);
  gtk_widget_show (ok_buttonFileDialog);
  GTK_WIDGET_SET_FLAGS (ok_buttonFileDialog, GTK_CAN_DEFAULT);

  /* Create the cancel button */
  cancel_buttonFileDialog = GTK_FILE_SELECTION (open_filesel)->cancel_button;
  /* Set the data for the cancel button widget so we can look it up easily */
  gtk_object_set_data (GTK_OBJECT (open_filesel), "cancel_buttonFileDialog", cancel_buttonFileDialog);
  gtk_widget_show (cancel_buttonFileDialog);
  GTK_WIDGET_SET_FLAGS (cancel_buttonFileDialog, GTK_CAN_DEFAULT);

  /* Connect the buttons to the event functions */
  gtk_signal_connect (GTK_OBJECT (ok_buttonFileDialog), "clicked", GTK_SIGNAL_FUNC (on_ok_buttonFileDialog_clicked), NULL);
  gtk_signal_connect (GTK_OBJECT (cancel_buttonFileDialog), "clicked", GTK_SIGNAL_FUNC (on_cancel_buttonFileDialog_clicked), NULL);
  /* Connect the cancel button to the file selection widget destroy, so the window will close */
  gtk_signal_connect_object(GTK_OBJECT (cancel_buttonFileDialog), "clicked", gtk_widget_destroy, GTK_OBJECT(open_filesel));
  /* Connect the ok button to the file selection widget destroy, so the window will close */
  gtk_signal_connect_object(GTK_OBJECT (ok_buttonFileDialog), "clicked", gtk_widget_destroy, GTK_OBJECT(open_filesel));

  /* Display the file selection dialog */
  gtk_widget_show (open_filesel);
  /* Raise it to the top window */
  gdk_window_raise (open_filesel->window);
  
}

void
on_ok_buttonFileDialog_clicked         (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *main_window, *des_text;
  gchar *out_file_txt, *config_file, *mpa2mka_label_txt;
  int opened_file;
  
  /* Extract filename from the file selection dialog */
  filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(open_filesel));

  /* Find out if we are getting the input file name or the output filename */
  if (filename_mode == "input")
  {
      main_window = lookup_widget (GTK_WIDGET (main_button), "main_window");
      des_text = lookup_widget (main_window, "entry_inputfile");
      /* Place the filename in the text box */
      gtk_entry_set_text (GTK_ENTRY(des_text), filename);
      des_text = lookup_widget (main_window, "entry_outputfile");
      if (strlen(gtk_entry_get_text(GTK_ENTRY(des_text))) < 1)
      {
        out_file_txt = merge_strings(filename, ".mka");
        /* Place the filename in the text box */
        gtk_entry_set_text (GTK_ENTRY(des_text), out_file_txt);
      }

  }
  if (filename_mode == "output")
  {
      main_window = lookup_widget (GTK_WIDGET (main_button), "main_window");
      des_text = lookup_widget (main_window, "entry_outputfile");
      /* Place the filename in the text box */
      gtk_entry_set_text (GTK_ENTRY(des_text), filename);
  }
  /* Reset the filename mode since we are done with it */
  filename_mode = "done";
}

void
on_cancel_buttonFileDialog_clicked     (GtkButton       *button,
                                        gpointer         user_data)
{
  /* This is already connected to the gtk_widget_destroy */
}


void
on_main_window_destroy                 (GtkObject       *object,
                                        gpointer         user_data)
{
  /* Exit the program completely */
  gtk_main_quit();
}

void show_simple_dialog (gchar *message_text, gchar *first_button_txt, gchar *second_button_txt)
{
  GtkWidget *dialog, *label, *first_button, *second_button;

  /* Create a new dialog box */
  dialog = gtk_dialog_new();
  gtk_window_set_transient_for (GTK_WINDOW(GTK_DIALOG(dialog)), GTK_WINDOW(main_button));
  gtk_window_set_title (GTK_WINDOW(GTK_DIALOG(dialog)), "Please read");
  gtk_window_set_modal (GTK_WINDOW(GTK_DIALOG(dialog)), TRUE);
  gtk_window_set_position (GTK_WINDOW(GTK_DIALOG(dialog)),GTK_WIN_POS_CENTER);
  /* Create the label widget of the dialog box */
  label = gtk_label_new (message_text);
   /* Create the first button */
  first_button = gtk_button_new_with_label(first_button_txt);
  if (first_button_txt == "Yes")
  {
    gtk_object_set_data (GTK_OBJECT (first_button), "mode", 1);
  }
  if (first_button_txt == "Write Debug Log")
  {
    gtk_object_set_data (GTK_OBJECT (first_button), "mode", 2);
  }
  /* Store widget pointer to ok button */
  //gtk_object_set_data_full (GTK_OBJECT (main_button), "buttonSimpleDialogButton1", first_button, (GtkDestroyNotify) gtk_widget_unref);
  /* Connect the first button to a signal handler function */
  gtk_signal_connect (GTK_OBJECT (first_button), "clicked", GTK_SIGNAL_FUNC (on_simple_dialog_button_clicked), NULL);
  /* Ensure that the dialog box is destroyed when the user clicks ok. */
  gtk_signal_connect_object(GTK_OBJECT (first_button), "clicked", gtk_widget_destroy, GTK_OBJECT(dialog));
  /* Add the first Button to the dialog  */
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->action_area), first_button);
  if (second_button_txt != "")
  {
    /* Create the second button */
    second_button = gtk_button_new_with_label(second_button_txt);
    /* Ensure that the dialog box is destroyed when the user clicks it. */
    gtk_signal_connect_object(GTK_OBJECT (second_button), "clicked", gtk_widget_destroy, GTK_OBJECT(dialog));
    /* Add the second Button to the dialog  */
    gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->action_area), second_button);
  };
  /* Add the label */
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG(dialog)->vbox), label);
  
  /* Show everything in the dialog. */
  gtk_widget_show_all (dialog);
}

gchar *merge_strings (gchar *first_string, gchar *second_string)
{
  /* Thanks to mosu for this little snippet of code */
  gchar *newstring;
  if ((strlen(first_string) + strlen(second_string)) >= 2) 
  {
    /* Allocate enough memory to hold both strings */
    newstring = (char *)malloc(strlen(first_string) + strlen(second_string) + 1);
    /* Copy the first string to a new string */
    strcpy(newstring, first_string);
    /* Copy the second string on the end of the new string */
    strcat(newstring, second_string);
  } 
  else
  {
    /* Strings are too short */
    newstring = "";
  }
  /* Return the new string */
  return newstring;
}

void
on_fileselection1_destroy              (GtkObject       *object,
                                        gpointer         user_data)
{
  /* This function is not needed */
}

void run_mpa2mka(gchar *source_file, gchar *des_file)
{
  /* For creation of command-line to run */
  GtkWidget *command_output_box, *cluster_time, *cluster_size;
  char *mpa2mka_command_line;
  int *console_output;
  gchar *standard_output_txt;
  char standard_output[1024] = "";
  int loop = 0;

  /* Create the command-line for launching  */
  mpa2mka_command_line = "mpa2mka ";
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, "-t ");
  cluster_time = lookup_widget(GTK_WIDGET(main_button), "entry_max_time");
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, gtk_entry_get_text (GTK_ENTRY(cluster_time)));
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, " -c ");
  cluster_size = lookup_widget(GTK_WIDGET(main_button), "entry_max_size");
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, gtk_entry_get_text (GTK_ENTRY(cluster_size)));
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, " \"");
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, source_file);
  mpa2mka_command_line = merge_strings(mpa2mka_command_line, "\"");
  if (des_file != "")
  {
    mpa2mka_command_line = merge_strings(mpa2mka_command_line, " \"");
    mpa2mka_command_line = merge_strings(mpa2mka_command_line, des_file);
    mpa2mka_command_line = merge_strings(mpa2mka_command_line, "\"");
  }
  else
  {
    mpa2mka_command_line = merge_strings(mpa2mka_command_line, " \"");
    mpa2mka_command_line = merge_strings(mpa2mka_command_line, merge_strings(source_file, ".mka"));
    mpa2mka_command_line = merge_strings(mpa2mka_command_line, "\"");
  }
  
  command_output_box = lookup_widget(GTK_WIDGET(main_button), "checkbutton_get_output");
  
  if (GTK_TOGGLE_BUTTON (command_output_box)->active)
  {
    /* With Pipe */
    console_output = popen (mpa2mka_command_line, "r");
    while (!feof(console_output))
    {
      standard_output[loop++] = fgetc(console_output);
      if (standard_output[loop-1] == 13)
      {
        loop--;
      }
    }
    pclose(console_output);
    
    command_output_box = lookup_widget(GTK_WIDGET(main_button), "text_output");
    /* Enable Editing */
    gtk_editable_set_editable (GTK_EDITABLE(command_output_box), TRUE);
    /* Delete any text already in the text box */
    gtk_editable_delete_text (GTK_EDITABLE (command_output_box), 0, -1);
    gtk_text_insert (command_output_box, NULL, NULL, NULL, "Commandline: ", -1);
    gtk_text_insert (command_output_box, NULL, NULL, NULL, mpa2mka_command_line, -1);
    gtk_text_insert (command_output_box, NULL, NULL, NULL, "\nResult:\n", -1);
    /* Convert console output to UTF-8 */
    standard_output_txt = g_locale_to_utf8 (standard_output, -1, NULL, NULL, NULL);
    if (standard_output_txt == NULL)
    {
    gtk_text_insert (command_output_box, NULL, NULL, NULL, standard_output, -1);
    }
    else
    {
    gtk_text_insert (command_output_box, NULL, NULL, NULL, standard_output_txt, -1);
    }  
  }
  else
  {
    /* Without Pipes */
    system(mpa2mka_command_line);
    command_output_box = lookup_widget(GTK_WIDGET(main_button), "text_output");
    /* Enable Editing */
    gtk_editable_set_editable (GTK_EDITABLE(command_output_box), TRUE);
    /* Delete any text already in the text box */
    //gtk_editable_delete_text (GTK_EDITABLE (command_output_box), 0, -1);
    gtk_text_insert (command_output_box, NULL, NULL, NULL, "Commandline: ", -1);
    gtk_text_insert (command_output_box, NULL, NULL, NULL, mpa2mka_command_line, -1);
  }
};

void on_simple_dialog_button_clicked (GtkButton *button, gpointer user_data)
{
  GtkWidget *in_text, *out_text;
  gchar *input_txt, *output_txt;
  gint dialog_choice;

  /* The first button in the dialog was clicked */
  dialog_choice = gtk_object_get_data (GTK_OBJECT (button), "mode");
  /* Yes */
  if (dialog_choice == 1)
  {
    /* Lookup the text box widgets */
    in_text = lookup_widget (GTK_WIDGET(main_button), "entry_inputfile");
    out_text = lookup_widget (GTK_WIDGET(main_button), "entry_outputfile");
    /* Retrieve the text from both text box widgets */
    input_txt = gtk_entry_get_text (GTK_ENTRY(in_text));
    output_txt = gtk_entry_get_text (GTK_ENTRY(out_text));
    run_mpa2mka(input_txt, output_txt);
  }
  /* Write Debug Log */
  if (dialog_choice == 2)
  {
  
  }
}
