/*
    Copyright (C) 2003, John Cannon <spyder482 @ yahoo.com>
    Copyright (C) 2003, Jory Stone <jcsston @ toughguy.net>
    Copyright (C) 2002, Steve Lhomme <steve.lhomme @ free.fr>
    This file is part of mpa2mka.

    Mpa2mka is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Mpa2mka is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
    \file mpa2mka.cpp
    Convert MPEG I/II Layer 1/2/3 to Matroska files
    \todo remove bad frames    
*/
#include "mpa2mka.h"

using namespace LIBMATROSKA_NAMESPACE;
using namespace LIBEBML_NAMESPACE;
using namespace std;

#if defined wxUSE_GUI
#define printf OutputToBuffer
void MPA2MKA::ConvertMPA2Matroska_wxWindows(wxTextCtrl *target_console_output_box, std::string passed_InputFile, std::string passed_OutputFile)
{
	console_output_box = target_console_output_box;
	ConvertMPA2Matroska(passed_InputFile, passed_OutputFile);
	return;
}
#endif

MPA2MKA::MPA2MKA()
{
	currentBlockGroup = NULL;
	clust = NULL;
	data = NULL;

	currentBlockGroupFull = false;
	noLacing = false;
	bVerbose = false;
	bCopyTags = true;
	maxTimePerCluster = 5000; //10 seconds
	maxClusterSize = 512 * 1024; //512kB
	currentClusterSize = 0;
	currentClusterBaseTimecode = 0;
	TIMECODE_SCALE = 1000000;
};

MPA2MKA::~MPA2MKA()
{
	//Free memory here
	//I need to know how to free memory used by libmatroska objects
	//delete frame_buffer;
	delete data;
};

bool MPA2MKA::SetInputFile(std::string passed_InputFile)
{
	input_Filename = passed_InputFile;
	return true;
};

bool MPA2MKA::SetOutputFile(std::string passed_OutputFile)
{
	output_Filename = passed_OutputFile;
	return true;
};

bool MPA2MKA::SetVerboseModeOn(bool bVerboseOn)
{
	bVerbose = bVerboseOn;
	return true;
};

bool MPA2MKA::SetLacingOff(bool bLacingOff)
{
	noLacing = bLacingOff;
	return true;
};

bool MPA2MKA::SetTagCopying(bool bTagCopying)
{
	bCopyTags = bTagCopying;
	return true;
}

bool MPA2MKA::SetMaxClusterByteSize(uint64 passed_maxClusterSize)
{
	//The allowed range is 128 bytes to 1073741824 bytes or 1GB
	if ((passed_maxClusterSize > 128) && (passed_maxClusterSize < 1073741824)) {
		maxClusterSize = passed_maxClusterSize;
		return true;
	}else {
		//Assign the default value
		maxClusterSize = 512 * 1024;
		return false;
	}
};

bool MPA2MKA::SetMaxClusterTimeSize(uint32 passed_maxTimePerCluster)
{
	//The allowed range is 10ms to 3600000ms or 1 hour
	if ((passed_maxTimePerCluster > 10) && (passed_maxTimePerCluster < 3600000)) {
		maxTimePerCluster = passed_maxTimePerCluster;
		return true;
	}else {
		//Assign the default value, 5 seconds
		maxTimePerCluster = 5000;
		return false;
	}
};

void MPA2MKA::OutputToBuffer(char *formating_string, ...)
{
	va_list argptr;

	//Get a pointer to the list of arg
	va_start(argptr, formating_string);	
#if defined wxUSE_GUI 
	//Redirect the output to a wxWindow wxTextCtrl
	wxString formated_text;
	formated_text.PrintfV(wxString(formating_string, wxConvUTF8).c_str(), argptr);
	console_output_box->AppendText(formated_text);
	//Give the processor some relief
	wxYield();
#else
	//Just let the function call pass through unaffected
	vprintf(formating_string, argptr);	
#endif
}

int MPA2MKA::ConvertMPA2Matroska(std::string passed_InputFile, std::string passed_OutputFile)
{
	using namespace M2W_NAMESPACE;
	//Store values
	input_Filename = passed_InputFile;

	if (passed_OutputFile == "")
		output_Filename = passed_InputFile + ".mka";
	else
		output_Filename = passed_OutputFile;

	try{
		printf("%s - (c) 2003 John Cannon & Jory Stone\n", mpa2mka_app_name);
		printf("Portions (c) 2002 Steve Lhomme\n");
		printf("Under the GNU GPL License\n\n");
		
		if (bVerbose)
			printf("In Verbose Mode, prepare to be blasted with numbers\n");
		if (noLacing)
			printf("Lacing turned off\n");

    // open the input file
    MPA_File input_File(input_Filename, MODE_READ);
    
    StdIOCallback output_file(output_Filename.c_str(), MODE_CREATE);

	
    // prepare to read the whole file for statistics
  	if (bVerbose){
      printf("Reading statistics in %s\n", input_Filename.c_str());
	  }

    input_File.SkipToFrameStart();

    MPA_Lame_Frame infoFrame;
    bool bInfoFrameOn = false;
    if (input_File.CheckInfoFrame()) {
		  *static_cast<MPA_Frame *>(&infoFrame) = input_File.GetInfoFrame();
		  bInfoFrameOn = true;
    }

    uint32 FrameNumber = 0;
    uint16 Layer = 0;
    uint64 SampleNumbers = 0;
    uint64 TotalFrameSize = 0;
    uint16 SamplingRate = 0;
    uint16 Channels = 0;
    uint16 SampleSize = 0;
    MPA_Frame frame;
    
    while (input_File.FindNextSyncWord()) {
		  FrameNumber++;
		  frame = input_File.GetFrame();
		  /// \todo handle the bitrate of the frame in a key/value table, so that we can detect real CBR files
		  TotalFrameSize += frame.DataSize() + 4;
		  SampleNumbers += frame.SampleSize();
		  SamplingRate = frame.SamplingRate();
		  Channels = frame.Channels();
		  SampleSize = frame.SampleSize();
		  Layer = frame.Layer();
		  input_File.SkipData();
    }

    uint64 totalByteSize = TotalFrameSize * SamplingRate;
    uint32 AverageBitrate, AverageByterate;
    if ((8 * totalByteSize) % SampleNumbers > SampleNumbers>>1)
	    AverageBitrate = 1 + 8 * totalByteSize / SampleNumbers;
    else
	    AverageBitrate = 8 * totalByteSize / SampleNumbers;

    if (AverageBitrate % 8 > 4)
	    AverageByterate = 1 + (AverageBitrate >> 3);
    else
	    AverageByterate = AverageBitrate >> 3;

	  printf("Maximum Cluster size is %dkB\n", (maxClusterSize/1024));
	  printf("Maximum time per cluster is %dms\n", maxTimePerCluster);

		Duration_total = (((int64)SampleNumbers * 1000) / SamplingRate);
    InitMatroskaOutputFile(output_file, SamplingRate, Layer, Channels);
    
    ChangeClusters(output_file, 0);
		currentBlockGroup  = &clust->GetNewBlock();
    
    // prepare to read for output
    input_File.SkipToFrameStart();

    /// \todo add the option to skip the Info frame or not
	  if (bInfoFrameOn) {
		  input_File.FindNextSyncWord();
		  input_File.GetFrame();
		  input_File.SkipData();
    }
    
	  if (bVerbose) {	
		  printf("Writing frames in %s\n", passed_OutputFile.c_str());
	  }
    uint32 prevSamplesNumber = 0;
    SampleNumbers = 0;
    int i=0;
    while (input_File.FindNextSyncWord()) {
		  //binary frame_buffer[5000+4]; // should be enough for 32kHz/320kbps
      frame_buffer = new binary[5000+4];
	    frame = input_File.GetFrame();
	    SampleNumbers += frame.SampleSize();
	    memcpy(frame_buffer, (const binary *)(frame), 4);
	    input_File.GetData(frame_buffer+4, 5000);
	    /// \todo handle timecode
      if (bVerbose) {
        printf("frame %4d : samples %d ", i++, SampleNumbers);
        printf("in pos %d ", input_File.getFilePointer());
        printf("size %d\n", frame.DataSize()+4);
		  }
		  AddFrame(output_file, frame_buffer, frame.DataSize()+4, (((uint64)prevSamplesNumber * 1000) / SamplingRate));
		  currentClusterSize += frame.DataSize()+4;		
		  prevSamplesNumber = SampleNumbers;
#if defined wxUSE_GUI
			//Give the processor some relief
			wxYield();
#endif			
    }	
    CloseMatroskaFile(output_file);

    output_file.close();
    input_File.close();

		currentBlockGroupFull = false;
		currentClusterSize = 0;
		currentClusterBaseTimecode = 0;

		printf("\n");
    return 0;    
    }catch (exception & Ex){
        cout << Ex.what() << endl;
        return -1;
    }
}

int MPA2MKA::InitMatroskaOutputFile(StdIOCallback & output_file, uint16 SamplingRate, uint16 Layer, uint16 Channels){
	try{
		// write output file
		EbmlHead FileHead;
	  
		EDocType & MyDocType = GetChild<EDocType>(FileHead);
		*static_cast<EbmlString *>(&MyDocType) = "matroska";
	  
		EDocTypeVersion & MyDocTypeVer = GetChild<EDocTypeVersion>(FileHead);
		*(static_cast<EbmlUInteger *>(&MyDocTypeVer)) = 1;
	  
		EDocTypeReadVersion & MyDocTypeReadVer = GetChild<EDocTypeReadVersion>(FileHead);
		*(static_cast<EbmlUInteger *>(&MyDocTypeReadVer)) = 1;
	  
		FileHead.Render(output_file);
	  
	  
		uint64 SegmentSize = FileSegment.Render(output_file);

		KaxTags & MyKaxTags = GetChild<KaxTags>(FileSegment);
		KaxTag & MyKaxTag = GetChild<KaxTag>(MyKaxTags);
		//using namespace id3lib; 
		if (bCopyTags) {
			//Start ID3 Tag preserving
			ID3_Tag tag;
			char tag_txt[1024];
			//unicode_t *tag_txt;
			
			//tag = ID3Tag_New();
			tag.Link(input_Filename.c_str());
			printf("\nID3 Tag Infomation: ");
			if (tag.HasTagType(ID3TT_ID3V1))
				printf("ID3v1 detected, ");
			if (tag.HasTagType(ID3TT_ID3V2))
				printf("ID3v2 detected, ");
			if (tag.HasTagType(ID3TT_MUSICMATCH))
				printf("MusicMatch ID3 tag detected, ");
			if (tag.HasTagType(ID3TT_LYRICS) | tag.HasTagType(ID3TT_LYRICS3) | tag.HasTagType(ID3TT_LYRICS3V2))
				printf("Lyrics tag detected, ");

			printf("%i fields found\n", tag.NumFrames());
			// use an std::auto_ptr here to handle object cleanup automatically
			ID3_Tag::Iterator* iter = tag.CreateIterator();
			ID3_Frame* myFrame = NULL;
			while (NULL != (myFrame = iter->GetNext()))
			{
				ID3_Field* myField = myFrame->GetField(ID3FN_TEXT);
				while (NULL != myField)
				{
					myField->Get(tag_txt, 1024);
					switch (myFrame->GetID())
					{
						case 1:
							break;
						case ID3FID_COMMENT:
						{
							//Create a new Multi-Identifier element
							KaxTagMultiComment & MyKaxTagMultiComment = GetChild<KaxTagMultiComment>(MyKaxTag);
							//Set the correct Language, English for now
							KaxTagMultiCommentLanguage & MyKaxTagMultiCommentLanguage = GetChild<KaxTagMultiCommentLanguage>(MyKaxTagMultiComment);
							*static_cast<EbmlString *>(&MyKaxTagMultiCommentLanguage) = "eng";
							//Fill the Name sub-element
							KaxTagMultiCommentComments & MyKaxTagMultiCommentComments = GetChild<KaxTagMultiCommentComments>(MyKaxTagMultiComment);
							*static_cast<EbmlUnicodeString *>(&MyKaxTagMultiCommentComments) = tag_txt;
							//Go ahead and render this tag
							//MyKaxTagMultiComment.Render(output_file);
							break;
						}
						case ID3FID_CDID:
						{
							//Create a new Multi-Identifier element
							KaxTagMultiIdentifier & MyKaxTagMultiIdentifier = GetChild<KaxTagMultiIdentifier>(MyKaxTag);
							//Set the correct Type
							KaxTagMultiIdentifierType & MyKaxTagMultiIdentifierType = GetChild<KaxTagMultiIdentifierType>(MyKaxTagMultiIdentifier);
							*static_cast<EbmlUInteger *>(&MyKaxTagMultiIdentifierType) = KaxTagMultiIdentifierType_CDIdentifier;
							//Fill the Name sub-element
							KaxTagMultiIdentifierString & MyKaxTagMultiIdentifierString = GetChild<KaxTagMultiIdentifierString>(MyKaxTagMultiIdentifier);
							*static_cast<EbmlUnicodeString *>(&MyKaxTagMultiIdentifierString) = tag_txt;
							//Go ahead and render this tag
							//MyKaxTagMultiIdentifier.Render(output_file);
							break;
						}
						case ID3FID_TITLE:
						{
							//Store the ID3 Title
							KaxTagMultiTitle & MyKaxTagMultiTitle = GetChild<KaxTagMultiTitle>(MyKaxTag);

							KaxTagMultiTitleType & MyKaxTagMultiTitleType = GetChild<KaxTagMultiTitleType>(MyKaxTagMultiTitle);
							*static_cast<EbmlUInteger *>(&MyKaxTagMultiTitleType) = KaxTagMultiTitleType_TrackTitle;

							KaxTagMultiTitleName & MyKaxTagMultiTitleName = GetChild<KaxTagMultiTitleName>(MyKaxTagMultiTitle);
							*static_cast<EbmlUnicodeString *>(&MyKaxTagMultiTitleName) = tag_txt;
							//MyKaxTagMultiTitle.Render(output_file);
							break;
						}
						case ID3FID_LEADARTIST:
						{
							//Store the ID3 Tag in the mka
							KaxTagMultiEntity & MyKaxTagMultiEntity = GetChild<KaxTagMultiEntity>(MyKaxTag);

							KaxTagMultiEntityType & MyKaxTagMultiEntityType = GetChild<KaxTagMultiEntityType>(MyKaxTagMultiEntity);
							*static_cast<EbmlUInteger *>(&MyKaxTagMultiEntityType) = KaxTagMultiEntitiesType_LeadPerformerSoloist;

							KaxTagMultiEntityName & MyKaxTagMultiEntityName = GetChild<KaxTagMultiEntityName>(MyKaxTagMultiEntity);
							*static_cast<EbmlUnicodeString *>(&MyKaxTagMultiEntityName) = tag_txt;
							//MyKaxTagMultiEntity.Render(output_file);
							break;
						}
						case ID3FID_CONTENTTYPE:
						{
							//Store the ID3 Genre
							KaxTagAudioGenre & MyKaxTagAudioGenre = GetChild<KaxTagAudioGenre>(MyKaxTag);
							*static_cast<EbmlString *>(&MyKaxTagAudioGenre) = tag_txt;
							//MyKaxTagAudioGenre.Render(output_file);
							break;
						}
						case ID3FID_TRACKNUM:
						{
							//Store the ID3 Disc Track Number
							KaxTagDiscTrack & MyKaxTagDiscTrack = GetChild<KaxTagDiscTrack>(MyKaxTag);
							*static_cast<EbmlUInteger *>(&MyKaxTagDiscTrack) = atol(tag_txt);
							//MyKaxTagDiscTrack.Render(output_file);
							break;
						}
						case ID3FID_WWWUSER:
						{
							//Store the ID3 Web URL
							KaxTagUserDefinedURL & MyKaxTagUserDefinedURL = GetChild<KaxTagUserDefinedURL>(MyKaxTag);
							*static_cast<EbmlString *>(&MyKaxTagUserDefinedURL) = tag_txt;
							//MyKaxTagUserDefinedURL.Render(output_file);
							break;
						}
						default:
							break;
					}
					//Clear the field so We'll look for the next frame
					myField = NULL;
				}
			}
			delete iter;
		}
		KaxTagFile & MyKaxTagFile = GetChild<KaxTagFile>(MyKaxTag);
		*static_cast<EbmlUnicodeString *>(&MyKaxTagFile) = input_Filename.c_str();

		//Fill Tag UID
		srand((unsigned)time(NULL));
		uint32 track_uid = rand();
		KaxTagTrackUID & MyKaxTagTrackUID = GetChild<KaxTagTrackUID>(MyKaxTag);
		*(static_cast<EbmlUInteger *>(&MyKaxTagTrackUID)) = track_uid;

		//Add info tag for writing app
		KaxWritingApp & MyKaxWritingApp = GetChild<KaxWritingApp>(MyKaxTags);
		*static_cast<EbmlUnicodeString *>(&MyKaxWritingApp) = mpa2mka_app_name;

		KaxMuxingApp & MyKaxMuxingApp = GetChild<KaxMuxingApp>(MyKaxTags);
		*static_cast<EbmlUnicodeString *>(&MyKaxMuxingApp) = "libmatroska";

		MyKaxTags.Render(output_file);

		//Add the Info segment
		KaxInfo & MyInfos = GetChild<KaxInfo>(FileSegment);
		KaxSegmentFilename & MyKaxSegmentFilename = GetChild<KaxSegmentFilename>(MyInfos);
		*static_cast<EbmlUnicodeString *>(&MyKaxSegmentFilename ) = input_Filename.c_str();

		//Add timecode scale
		KaxTimecodeScale & TimeScale = GetChild<KaxTimecodeScale>(MyInfos);
		*(static_cast<EbmlUInteger *>(&TimeScale)) = TIMECODE_SCALE;

		KaxDuration & Duration = GetChild<KaxDuration>(MyInfos);
		*(static_cast<EbmlFloat *>(&Duration)) = Duration_total;

		MyInfos.Render(output_file);

		KaxTracks & MyTracks = GetChild<KaxTracks>(FileSegment);

		//Fill in track params
		MyTrack1 = &GetChild<KaxTrackEntry>(MyTracks);
		KaxTrackNumber & MyTrack1Number = GetChild<KaxTrackNumber>(*MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyTrack1Number)) = 1;
		*(static_cast<EbmlUInteger *>(&GetChild<KaxTrackType>(*MyTrack1))) = track_audio;
		
		KaxTrackUID & MyKaxTrackUID = GetChild<KaxTrackUID>(*MyTrack1);
		*(static_cast<EbmlUInteger *>(&MyKaxTrackUID)) = track_uid;

		KaxCodecID & MyTrack1CodecID = GetChild<KaxCodecID>(*MyTrack1);
		if(Layer == 2){//MP3
				MyTrack1CodecID.CopyBuffer((binary*)"A_MPEG/L3",countof("A_MPEG/L3"));
		}else if(Layer == 1){//MP2
				MyTrack1CodecID.CopyBuffer((binary*)"A_MPEG/L2",countof("A_MPEG/L2"));
		}else{ //MP1 or unknown    
				MyTrack1CodecID.CopyBuffer((binary*)"A_MPEG/L1",countof("A_MPEG/L1"));    
		}

		KaxTrackAudio & MyTrack1Audio = GetChild<KaxTrackAudio>(*MyTrack1);

		KaxAudioSamplingFreq & MyTrack1Freq = GetChild<KaxAudioSamplingFreq>(MyTrack1Audio);
		*(static_cast<EbmlFloat *>(&MyTrack1Freq)) = (float)SamplingRate;

		MyTrack1Freq.ValidateSize();

		KaxAudioChannels & MyTrack1Channels = GetChild<KaxAudioChannels>(MyTrack1Audio);
		*(static_cast<EbmlUInteger *>(&MyTrack1Channels)) = Channels;

		uint64 TrackSize = MyTracks.Render(output_file);
		printf("Matroska File created...\n");
		return true;
	}catch (exception & Ex)
	{
		cout << Ex.what() << endl;        
	}
	return false;
}

bool MPA2MKA::AddFrame(StdIOCallback & output_file, binary* buffer, uint64 bufferLength, uint64 timecode){
  try{       
		bool clusterChange = ChangeClusters(output_file, timecode);
    if((!currentBlockGroup) || currentBlockGroupFull|| clusterChange) {        
      //AllCues.AddBlockGroup(*currentBlockGroup);    
			currentBlockGroup = &clust->GetNewBlock();
    }            

		data = new SimpleDataBuffer(buffer, bufferLength, 0);    
		if (noLacing) {
			currentBlockGroupFull = currentBlockGroup->AddFrame(*MyTrack1,(timecode-currentClusterBaseTimecode), *data);
		}else {
			currentBlockGroupFull = !currentBlockGroup->AddFrame(*MyTrack1,(timecode-currentClusterBaseTimecode), *data);
		}
  }catch (exception & Ex) {
    cout << Ex.what() << endl;
    return false;
  }
  return true;    
}

bool MPA2MKA::ChangeClusters(StdIOCallback & output_file, uint64 timecode){
  try{
    if(!clust || ((timecode - currentClusterBaseTimecode) >= maxTimePerCluster) || (currentClusterSize >= maxClusterSize)){        
      if(clust){
        clust->Render(output_file, AllCues);
        clust->ReleaseFrames();
        delete clust;
				delete data;
				//delete frame_buffer;
        currentClusterSize = 0;
      }
      clust = new KaxCluster();
      clust->SetParent(FileSegment);			
      clust->SetPreviousTimecode(currentClusterBaseTimecode);
      KaxClusterTimecode & MyClusterTimeCode = GetChild<KaxClusterTimecode>(*clust);
      *(static_cast<EbmlUInteger *>(&MyClusterTimeCode)) = timecode;   
      currentClusterBaseTimecode = timecode;
      return true;
    }else{
      return false;
    }
  }catch (exception & Ex){
    cout << Ex.what() << endl;
    return false;
  }   
}

void MPA2MKA::CloseMatroskaFile(StdIOCallback & output_file){
	if(clust){
    clust->Render(output_file, AllCues);
    clust->ReleaseFrames();
    delete clust;		
  }
  printf("File closed.\n");
}
